# Configuration file for the Sphinx documentation builder.

import os
from importlib import metadata

# -- Project information -----------------------------------------------------

project = "abaco"
copyright = "2025, Edir Vidal and MoNA group"
author = "Edir Vidal, Angel L.P., Henry Webel, Atieh Gharib, Juliana Assis, Sebastián Ayala-Ruano, Alberto Santos"
PACKAGE_VERSION = metadata.version("abaco")
version = PACKAGE_VERSION
release = PACKAGE_VERSION


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",  # Core extension for generating documentation from docstrings
    "sphinx.ext.autodoc.typehints",  # Automatically document type hints in function signatures
    "sphinx.ext.viewcode",  # Include links to the source code in the documentation
    "sphinx.ext.napoleon",  # Support for Google and NumPy style docstrings
    "sphinx.ext.intersphinx",  # allows linking to other projects' documentation in API
    "sphinx_new_tab_link",  # each link opens in a new tab
    "myst_nb",  # Markdown and Jupyter Notebook support
    "sphinx_copybutton",  # add copy button to code blocks
]

#  https://myst-nb.readthedocs.io/en/latest/computation/execute.html
nb_execution_mode = "auto"
nb_execution_timeout = -1  # -1 means no timeout
myst_enable_extensions = ["dollarmath", "amsmath"]

# Plotly support through require javascript library
# https://myst-nb.readthedocs.io/en/latest/render/interactive.html#plotly
html_js_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"
]

# https://myst-nb.readthedocs.io/en/latest/configuration.html
# Execution
nb_execution_raise_on_error = True
# Rendering
nb_merge_streams = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Ignore
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "jupyter_execute",
    "conf.py",
]

# Intersphinx options
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "scikit-learn": ("https://scikit-learn.org/stable/", None),
    "torch": ("https://docs.pytorch.org/docs/stable/", None),
    "transformers": ("https://huggingface.co/docs/transformers/main/en/", None),
}

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_book_theme"
html_logo = "https://raw.githubusercontent.com/Multiomics-Analytics-Group/abaco/refs/heads/main/docs/figures/overview.png"
html_favicon = "https://raw.githubusercontent.com/Multiomics-Analytics-Group/abaco/refs/heads/main/docs/figures/overview.png"
html_theme_options = {
    "github_url": "https://github.com/Multiomics-Analytics-Group/abaco",
    "repository_url": "https://github.com/Multiomics-Analytics-Group/abaco",
    "repository_branch": "main",
    "home_page_in_toc": True,
    "path_to_docs": "docs",
    "show_navbar_depth": 1,
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_download_button": True,
    "launch_buttons": {
        "colab_url": "https://colab.research.google.com"
    },
    "navigation_with_keys": False,
}

# -- Setup for sphinx-apidoc -------------------------------------------------

if os.environ.get("READTHEDOCS") == "True":
    from pathlib import Path

    PROJECT_ROOT = Path(__file__).parent.parent
    PACKAGE_ROOT = PROJECT_ROOT / "src" / "abaco"

    def run_apidoc(_):
        from sphinx.ext import apidoc

        apidoc.main(
            [
                "--force",
                "--implicit-namespaces",
                "--module-first",
                "--separate",
                "-o",
                str(PROJECT_ROOT / "docs" / "reference"),
                str(PACKAGE_ROOT),
                str(PACKAGE_ROOT / "*.c"),
                str(PACKAGE_ROOT / "*.so"),
            ]
        )

    def setup(app):
        app.connect("builder-inited", run_apidoc)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
#   kernelspec:
#     display_name: .venv
#     language: python
#     name: python3
# ---

# %% [markdown]
# # `ABaCo` tutorial: Anaerobic Digestion benchmark dataset

# %% [markdown]
# In this notebook we will implement ABaCo for batch correction on an Anaerobic Digestion (AD) dataset under different phenol concentration conditions. 
#
# -----
# **Data Description:**
# - The AD dataset generated by ([Chapleur et al., 2016](https://doi.org/10.1007/s10532-015-9751-4)) and shared in R package [PLSDAbatch](https://github.com/EvaYiwenWang/PLSDAbatch) by [Wang and Cao (2023)](https://doi.org/10.1093/bib/bbac622) is a dataset composed of 75 samples with 567 identified taxonomic groups. 
# - The samples were treated with two different phenol concentrations, accounting for the biological source of variation. 
# - Samples were processed on 5 different dates over the lapse of 2 years, accounting for the technical source of variation (i.e., batch effect).
#
# **Goal:**
#
# With ABaCo the aim is to remove the technical variation (batch effect) while retaining the biological variation (phenol group effect). 
#
# -----

# %% [markdown]
# ## Dataset Requirements for ABaCo
#
# The dataset should contain the following to be compatible with the ABaCo framework:
#
# | sample | batch      | trt  | feat1 | feat2 | ... |
# |--------|------------|------|-------|-------|-----|
# | A      | 24/07/2025 | low  | #     | #     | ... |
# | B      | 15/06/2024 | low  | #     | #     | ... |
# | C      | 24/07/2025 | high | #     | #     | ... |
#
# - The data must have 3 categorical columns: 
#     1. unique ids to identify the observations/samples e.g. sample ids `sample`
#     2. ids for the batch/factor groupings to be corrected by abaco. e.g. dates of sample analysis `batch`
#     3. biological/experimental factor variation for abaco to retain when correcting batch effect e.g., phenol concentration condition `trt`
#
# - And the features (numeric type) to be trained on. 
#
# `abaco` provides a `BatchEffectDataLoader.dataPreprocess()` function to help convert a plain text file (e.g, csv, tsv) into a compatible pd.DataFrame format, and a `onehotencoding()` function to encode categorical columns. 

# %%
from abaco.dataloader import DataPreprocess, one_hot_encoding

# Load AD count
path_to_dataset = "data/dataset_ad.csv"
batch_col = "batch"
id_col = "sample"
bio_col = "trt"

# Convert data path into compatible pd.DataFrame with ABaCo framework
df_ad = DataPreprocess(
    path_to_dataset,
    factors=[
        batch_col, 
        id_col, 
        bio_col
    ]
)

# see that there are 3 categorical and n numeric columns
print(df_ad.info())

# Isolating the features (numeric) to be inputed into the model
df_features = df_ad.drop(columns=[batch_col, id_col, bio_col]).values
print(f"\n The feature inputs have shape: {df_features.shape}")

# %%
# One-hot encode the categorical columns
# the batch factor
ohe_batches, ohe_batch_classes = one_hot_encoding(df_ad[batch_col])
# and the biological factor
ohe_bio, ohe_bio_classes = one_hot_encoding(df_ad[bio_col])

print(f"Batch classes: {ohe_batch_classes}\n")
print(f"Biological classes: {ohe_bio_classes}")

# %% [markdown]
# ### The AD dataset in brief
# `abaco` also provides plotting functions to easily explore the data. Here we use `BatchEffectPlots.plotPCoA()` to visualize the batch and biological effects on the data.

# %%
from abaco.plots import plotPCoA
import plotly.io as pio
pio.renderers.default = "notebook_connected"
# pcoa
plotPCoA(
    data=df_ad, 
    sample_label=id_col, 
    batch_label=batch_col, 
    experiment_label=bio_col,
)

# %% [markdown]
# - Batch effect (colours): 
#     - Samples were processed on 5 different dates over the lapse of 2 years, accounting for the technical source of variation 
#     - The points clustering by colour demonstrate that there is a batch effect.
# - Biological effect (shapes): 
#     - The samples were treated with 2 different phenol concentrations, accounting for the biological source of variation. 
#     - Within the batches the circle points (lower phenol conc.) are to the left of the square points (higher conc).
#
#
# ### The goal 
#
# The aim of **ABaCo** is to: 
# 1) correct the batch effect (e.g., the points should no longer cluster by colour in the PCoA) while
# 2) maintaining biological variance (e.g., keep circles and squares separated).
#
# Ideally, after using AbaCo to transform the data, the resulting PCoA will look like a colourful mixture of points, where the circle points remain to the left of the square ones.
#
#
# ---

# %% [markdown]
# # Using `ABaCo`

# %% [markdown]
# ## Training the ABaCo model
#
# To train ABaCo on the prepared AD dataset, we create the `abaco.metaABaCo()` class and pass the required parameters shown in the cell below. Thus, running abaco with the default optional parameters.
#
# Usually, setup of the parameters is required, which are explained in brief in the documentation e.g. `help(metaABaCo)`

# %%
from abaco.ABaCo import metaABaCo
import torch

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Create ABaCo model
model = metaABaCo(
    data=df_ad, # Pre-processed dataframe
    n_bios=2, # Number of biological groups in the data
    bio_label=bio_col, # Column where biological groups are labeled in the dataframe
    n_batches=5, # Number of batch groups in the data
    batch_label=batch_col, # Column where batch groups are labeled in the dataframe
    n_features=567, # Number of features (taxonomic groups)
    prior="MoG", # Prior distribution 
    device=device, # Device
)

# %% tags=["hide-output"]
help(metaABaCo)

# %% [markdown]
# For training the model we can call the method `metaABaCo.fit()` with the default parameters.

# %% tags=["hide-output"]
model.fit(seed=42)

# %% [markdown]
# After training the abaco model we can use it to reconstruct the AD dataset without the batch variance while keeping the biological variance.
#
# -----

# %% [markdown]
# ## ABaCo data reconstruction

# %% [markdown]
# To reconstruct the dataset we use the method `metaABaco.correct()`.

# %%
# Reconstruct the dataset using the trained ABaCo model
corrected_dataset = model.correct(seed=42)

# Plot the PCoA of the reconstructed dataset
plotPCoA(
    data = corrected_dataset, 
    sample_label=id_col, 
    batch_label=batch_col, 
    experiment_label=bio_col
)

# %% [markdown]
# ### Conclusion
#
# The goal was to: 
#
# &#x2705; correct the batch effect (reduce clustering of points by colour)
#
# &#x2705; maintain biological variance (maintain separation of circles and sequares)
#
# A brief visual inspection of the PCoA of the reconstructed AD data suggests that ABaCo reduced the batch effect associated with processing the samples on different days, while still retaining the variance due to the experimental condition of lower vs higher phenol concentration.
#
#
# ---



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
#   kernelspec:
#     display_name: .venv
#     language: python
#     name: python3
# ---

# %% [markdown]
# # `ABaCo` demo: Mgnify study `MGYS00001135`

# %% [markdown]
# In this demo we will implement ABaCo for batch correction on Mgnify study [MGYS00001135](https://www.ebi.ac.uk/metagenomics/studies/MGYS00001135#overview) and ENA Bioproject [PRJEB13832](https://www.ebi.ac.uk/ena/browser/view/PRJEB13832), a proof-of-concept study for surveilling infections diseases and AMR via metagenomics analysis of local sewage samples in Copenhagen.
#
# -----
# **Data Description:**
# - The metagenomics dataset collected by the Technical University of Denmark (DTU) and analysed via Mgnify [pipeline version 2.0](https://www.ebi.ac.uk/metagenomics/pipelines/2.0) is a dataset composed of 54 samples with 749 identified taxa at the genus level. 
# - The samples were collected from 3 different wastewater treatment plant (WWTP) locations, accounting for the biological source of variation. 
# - Samples were analyzed on 2 different dates accounting for the technical source of variation (i.e., batch effect).
# - The sample data and metadata were retrieved from the [Mgnify API](https://docs.mgnify.org/src/docs/api.html) using the following [mgnifyapi repo](https://github.com/Multiomics-Analytics-Group/Retrieve_info_MGnifyAPI/tree/angel-learning) and stored as an Annotated Dataset [AnnData](https://anndata.readthedocs.io/en/latest/index.html). 
#
# **Goal:**
#
# With ABaCo the aim is to remove the technical variation (batch effect) while maintaining any biological variation (wastewater treatment plant location). 
#
# -----
#
# To start we read in the annotated dataset. 

# %%
import anndata as ad
import pandas as pd 
import numpy as np

# read in annotated dataset
adata = ad.read_h5ad("data/mgnify_MGYS00001135.h5ad")
# print details
adata

# %% [markdown]
# The columns (variables, vars) are the different taxa annotated with taxonomic ranks (sk = super kingdom, k = kingdom, p = phylum, c = class, o = order, f = family, g = genus).
#
# The rows (observations, obs) are the metadata for the different samples in the study MGYS00001135. 
#
# The ones of interest for this demo are: 
# - The `analysis_completed_date` which accounts for batch effect for our purposes, and
# - `ww_plant` which are the different sampling locations and we will use as our biological variance that we wish to retain after batch correction.
#
# -----
#
# We can easily extract a df from anndata so that we meet the data format required for abaco.

# %%
batch_col = 'analysis_completed_date'
bio_col = 'ww_plant'
id_col = 'index'

# the raw counts as df
df_taxa = adata.to_df()

# appending on the categorical data of interest
df_all = pd.concat([df_taxa, adata.obs[[batch_col, bio_col]]], axis=1).reset_index()
df_all.info()

# %% [markdown]
# **`pd.DataFrame` Requirements for ABaCo:**
#
# The dataset contains the following making it compatible with the ABaCo framework:
#
# | id_col | batch_col  | bio_col  | count1 | count2 | ... |
# |--------|------------|----------|--------|--------|-----|
# | A      | 24/07/2025 | RA       | #      | #      | ... |
# | B      | 15/06/2024 | RD       | #      | #      | ... |
# | C      | 24/07/2025 | RL       | #      | #      | ... |
#
# - The data has categorical columns: 
#     1. unique ids to identify the observations/samples e.g. sample id col
#     2. ids for the batch/factor groupings to be corrected by abaco. e.g. dates of sample analysis batch_col
#     3. biological/experimental factor variation for abaco to retain when correcting batch effect e.g., sampling location bio_col
#
# - And taxa counts to be trained on. 
#
# We can use `abaco.plots.plotPCoA()` to visualize any batch and biological effects based on the given categories. 

# %%
from abaco.plots import plotPCoA
import plotly.io as pio
pio.renderers.default = "notebook_connected"

plotPCoA(
    data=df_all, 
    sample_label=id_col, 
    batch_label=batch_col,  
    experiment_label=bio_col,
)

# %% [markdown]
# - Batch effect (colours): 
#     - Analysis of the samples were completed on 2 different dates which could result in a technical source of variation captured by the slight clustering of batches along PCo1
# - Biological effect (shapes): 
#     - The samples were collected from 3 different wwtps which we use as the biological source of variation for abaco to retain during batch correction. 
#     - There isn't a clear clustering of shapes in the PCoA plot, but we can take a look at other metrics such as alpha diversity indices between wwtp locations that should be retained after abaco reconstruction.

# %% [markdown]
# ## Alpha diversity of original samples
#
# For each sample we will calculate the below richess, evenness and diversity indices and store them as additional sample (obs) metadata in the anndata set.
#
# Richness
# - number of unique species per sample
# - Chao1 estimate - correction to above which is likely an underestimation
#
# Diversity
# - Shannon diversity index - if a sample is more diverse then there is more entropy (predicting the next taxa is more difficult) and thus the Shannon diversity index is higher. 
# - Simpson index (and inverse) - probability of 2 microbes being the same taxa (Simpson index closer to 0 is more diverse, while a higher inverse Simpson index means a sample is more diverse)
#
# Evenness
# - Pielou's evenness index - when abundances are more homogenous in a sample then more even so higher Pielou
#
# The formulas are included with the below code. The metrics name will be prefixed with 'X_' so that we can later compare them to 'abaco_' metrics. 
#
# ### Preprocessing the count tables
#
# The raw counts require preprocessing because samples will have different total read counts due to technical differences during sequencing. We can normalize the samples to get relative abundances of the taxa summing to 1 (composition data).

# %%
# normalize original data and store as additional anndata layer
adata.layers['normalized_X'] = (adata.X/adata.X.sum(axis=1).reshape(-1,1))
# sanity check 
adata.layers['normalized_X'].sum(axis=1)

# %% [markdown]
# Our normalized, compositional data (CoDa) is bound to a simplex. To move forward with any statistical tests or ML we can choose to map the coda to unconstrained vectors by log-ratio transforming (e.g., centred log-ratio via `skbio.stats.composition.clr()`) the data. 

# %%
from skbio.stats.composition import clr 

# CLR transform normalized data, replacing zeros with smallll val to avoid log(0)
adata.layers['clr_X'] = clr(np.where(adata.layers['normalized_X'] > 0, adata.layers['normalized_X'], 1e-10))

# save anndata with new layers
adata.write_h5ad("data/mgnify_MGYS00001135.h5ad")

# %% [markdown]
# ### Alpha diversity as sample metadata

# %%
# Diversity: Number of unique taxa per sample as new metadata
adata.obs['X_numtaxa'] = (adata.X > 0).sum(axis=1) 

# Diversity with correction: Chao1 estimator
# Chao1 = num observed taxa + (num singletons / (2 x num doubletons))
num_singletons = (adata.X == 1).sum(axis=1)
num_doubletons = (adata.X == 2).sum(axis=1)
adata.obs['X_chao1'] = (adata.X > 0).sum(axis=1) + (num_singletons / (2 * num_doubletons))

# Diversity: Shannon index H = -sum(p_i * log(p_i)), entropy
from scipy.stats import entropy
adata.obs['X_shannon'] = entropy(adata.layers['normalized_X'], nan_policy='omit', axis=1)

# Evenness: Pielou evenness index P = H/H_max, H_max = ln(num species)
adata.obs['X_pielou'] = adata.obs['X_shannon'] / np.log(adata.obs['X_numtaxa'])

# Diversity: simpson index lambda = sum(p_i ^2)
adata.obs['X_simpson'] = (adata.layers['normalized_X'] ** 2).sum(axis=1)

# Diversity: inverse simpson index D = 1/lambda
adata.obs['X_inv_simpson'] = 1 / adata.obs['X_simpson']

# %% [markdown]
# We can quickly visualize the distribution and relationships between the metrics by batch and biological groups using [`seaborn.pairplot()`](https://seaborn.pydata.org/generated/seaborn.pairplot.html)

# %%
import seaborn as sns
import matplotlib.pyplot as plt

# preparing the df for vis
df_alphas = adata.obs[[
    'X_numtaxa', 'X_chao1', 'X_shannon', 
    'X_pielou', 'X_simpson', 'X_inv_simpson',
    batch_col, bio_col
]].copy()

# plotting by bio groups
g1 = sns.pairplot(df_alphas, hue=bio_col, corner=True, height=1.8)
g1.figure.suptitle('Alpha Diversity by Biological Groups', fontsize=22)
plt.setp(g1._legend.get_texts(), fontsize=18)
plt.setp(g1._legend.get_title(), fontsize=20)

g2 = sns.pairplot(df_alphas, hue=batch_col, corner=True, height=1.8)
g2.figure.suptitle('Alpha Diversity by Batch Groups', fontsize=22)
plt.setp(g2._legend.get_texts(), fontsize=18)
plt.setp(g2._legend.get_title(), fontsize=20)

# %% [markdown]
# The pairplots suggest that the batch effect could be influencing the alpha diversity of the samples. 
#
# -----

# %% [markdown]
#
#
# ## The goal 
#
# Here the aim of **ABaCo** is to: 
# 1) correct the batch effect (e.g., the points should no longer cluster by colour in the PCoA and pairplot) while
# 2) maintaining biological variance (e.g., the pairplot coloured by bio group should be not change).
#
# Ideally, after using AbaCo to transform the data, the resulting PCoA and pairplot coloured by batches will look like a colourful mixture of points and the pairplot coloured by wwtp location should not change.
#
# -----
#
# ## Using `ABaCo`
#
# ### Setting up ABaCo
#
# We instantiate the `abaco.metaABaCo()` class and pass the required parameters shown in the cell below. Thus, running abaco with the default optional parameters.
#
# Usually, setup of the parameters is required, which are explained in brief in the documentation e.g. `help(metaABaCo)`

# %%
from abaco.ABaCo import metaABaCo
import torch

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Instaantiate the ABaCo model
model = metaABaCo(
    data=df_all, # Pre-processed dataframe
    n_bios=df_all[bio_col].nunique(), # Number of biological groups in the data
    bio_label=bio_col, # Column where biological groups are labeled in the dataframe
    n_batches=df_all[batch_col].nunique(), # Number of batch groups in the data
    batch_label=batch_col, # Column where batch groups are labeled in the dataframe
    n_features=df_taxa.shape[1], # Number of features (taxonomic groups)
    prior="MoG", # Prior distribution 
    device=device, # Device
)

# %% tags=["hide-output"]
help(metaABaCo)

# %% [markdown]
# ### Training the ABaCo model
#
# To train ABaCo on the prepared dataset, we then use method `abaco.metaABaCo.fit()`. 

# %% tags=["hide-output"]
# Train the model
model.fit(seed=42)

# %% [markdown]
# ### Reconstructing the dataset with ABaCO
#
# To reconstruct the dataset we use method `abaco.metaABaCo.correct()` and save it as a new layer `abaco` in the annotated dataset. 

# %%
# Reconstruct the dataset using the trained ABaCo model
corrected_dataset = model.correct(seed=42)
# save back to adata
adata.layers['abaco'] = corrected_dataset.set_index(id_col).drop(columns=[batch_col, bio_col]).values
adata.write_h5ad("data/mgnify_MGYS00001135.h5ad")

# %% [markdown]
# ## Viewing the ABaCo reconstructed dataset
#
# Here we again create PCoA and look at sample alpha diversities via pairplots but on the batch corrected data.

# %%
# Plot the PCoA of the reconstructed dataset
plotPCoA(
    data = corrected_dataset, 
    sample_label=id_col, 
    batch_label=batch_col, 
    experiment_label=bio_col
)

# %% [markdown]
# In the PCoA we see that there is more overlap of the batches based on the blue and red points being more mixed. 
#
# We can also take a look at the alpha diversity metrics:

# %% tags=["hide-input"]
## PREPROCESSING THE ABaCo DATA
# normalize abaco corrected data
adata.layers['normalized_abaco'] = (adata.layers['abaco']/adata.layers['abaco'].sum(axis=1).reshape(-1,1))
# CLR transform normalized ABaCo data
adata.layers['clr_abaco'] = clr(np.where(adata.layers['normalized_abaco'] > 0, adata.layers['normalized_abaco'], 1e-10))

## Alpha diversity metrics on ABaCo corrected data
# Diversity: Number of unique taxa per sample as new metadata
adata.obs['abaco_numtaxa'] = (adata.layers['abaco'] > 0).sum(axis=1) 
# Diversity with correction: Chao1 estimator
# Chao1 = num observed taxa + (num singletons / (2 x num doubletons))
num_singletons = (adata.layers['abaco'] == 1).sum(axis=1)
num_doubletons = (adata.layers['abaco'] == 2).sum(axis=1)
adata.obs['abaco_chao1'] = (adata.layers['abaco'] > 0).sum(axis=1) + (num_singletons / (2 * num_doubletons))
# Diversity: Shannon index H = -sum(p_i * log(p_i)), entropy
adata.obs['abaco_shannon'] = entropy(adata.layers['normalized_abaco'], nan_policy='omit', axis=1)
# Evenness: Pielou evenness index P = H/H_max, H_max = ln(num species)
adata.obs['abaco_pielou'] = adata.obs['abaco_shannon'] / np.log(adata.obs['abaco_numtaxa'])
# Diversity: simpson index lambda = sum(p_i ^2)
adata.obs['abaco_simpson'] = (adata.layers['normalized_abaco'] ** 2).sum(axis=1)
# Diversity: inverse simpson index D = 1/lambda
adata.obs['abaco_inv_simpson'] = 1 / adata.obs['abaco_simpson']

## PAIRPLOTS OF ALPHA DIVERSITY METRICS
# prep df for vis
df_abaco_alphas = adata.obs[[
    'abaco_numtaxa', 'abaco_chao1', 'abaco_shannon', 
    'abaco_pielou', 'abaco_simpson', 'abaco_inv_simpson',
    batch_col, bio_col
]].copy()
# plotting by bio groups
g1 = sns.pairplot(df_abaco_alphas, hue=bio_col, corner=True, height=1.8)
g1.figure.suptitle('ABaCo: Alpha Diversity by Biological Groups', fontsize=22)
plt.setp(g1._legend.get_texts(), fontsize=18)
plt.setp(g1._legend.get_title(), fontsize=20)
g2 = sns.pairplot(df_abaco_alphas, hue=batch_col, corner=True, height=1.8)
g2.figure.suptitle('ABaCo: Alpha Diversity by Batch Groups', fontsize=22)
plt.setp(g2._legend.get_texts(), fontsize=18)
plt.setp(g2._legend.get_title(), fontsize=20)

# %% [markdown]
# ### Conclusion
#
# The goal was to: 
#
# &#x2705; correct the batch effect (reduce clustering of points by colour in the PCoA and pairplot coloured by batches)
#
# &#x2705; maintain biological variance (maintain pattern of shapes in PCoA and maintain pairplot coloured by wwtp location)
#
# A brief visual inspection of the PCoA and alpha diversity of the reconstructed data suggests that ABaCo reduced the batch effect associated with analyzing the samples on different dates, while still retaining any variance due to the condition of wwtp location.
#
#
# ---



# these functions are pretty general (file that can be reused across projects)
import argparse
import logging
import os
import sys
from datetime import datetime
from urllib.parse import urlparse, urlunsplit

# import yaml


## CHECKS
def assert_path(filepath: str):
    """
    Check that the given filepath is a string and that it exists.

    Parameters
    ----------
    filepath : str
        The filepath or folder path to check.

    Raises
    ------
    TypeError
        If the filepath is not a string.
    FileNotFoundError
        If the filepath does not exist.

    Example
    -------
    >>> assert_path("..")
    >>> assert_path("./tests")
    """
    if not isinstance(filepath, str):
        raise TypeError(f"filepath must be a string: {type(filepath)}")
    if not os.path.exists(os.path.abspath(filepath)):
        raise FileNotFoundError(f"The specified path does not exist: {filepath}")


def create_folder(directory_path: str, is_nested: bool = False) -> bool:
    """
    Create a folder if it doesn't exist.

    Parameters
    ----------
    directory_path : str
        The path of the directory to create.
    is_nested : bool, optional
        Whether to create nested directories (True uses os.makedirs, False uses os.mkdir), by default False.

    Returns
    -------
    bool
        True if the folder was created, False if it already existed.

    Raises
    ------
    TypeError
        If directory_path is not a string.
    ValueError
        If directory_path is an existing file.
    OSError
        If there is an error creating the directory.
    """
    # PRECONDITION CHECK
    if not isinstance(directory_path, str):
        raise TypeError(f"filepath must be a string: {type(directory_path)}")
    abs_path = os.path.abspath(directory_path)

    # make sure it is a folder not a file
    if os.path.isfile(abs_path):
        raise ValueError(
            f"directory_path is an existing file when it should be a folder/foldername: {abs_path}"
        )
    # if folder already exists
    elif os.path.isdir(abs_path):
        return False
    # create the folder(s)
    else:
        try:
            if is_nested:
                # Create the directory and any necessary parent directories
                os.makedirs(directory_path, exist_ok=True)
                return True
            else:
                # Create only the final directory (not nested)
                os.mkdir(directory_path)
                return True
        except OSError as e:
            raise OSError(f"Error creating directory '{directory_path}': {e}") from e


def assert_nonempty_keys(dictionary: dict):
    """
    Check that the keys in a dictionary are not empty strings.

    Parameters
    ----------
    dictionary : dict
        A dictionary (e.g., config file).

    Raises
    ------
    AssertionError
        If dictionary is not a dict or if any key is empty or blank.
    """
    # PRECONDITIONS
    if not isinstance(dictionary, dict):
        raise TypeError(f"dictionary must be a dict, got {type(dictionary)}")

    # MAIN FUNCTION
    for key in dictionary:
        if type(key) is str:
            assert key, f'There is an empty key (e.g., ""): {key, dictionary.keys()}'
            assert (
                key.strip()
            ), f'There is a blank key (e.g., space, " "): {key, dictionary.keys()}'


def assert_nonempty_vals(dictionary: dict):
    """
    Check that the values in a dictionary are not empty strings.

    Parameters
    ----------
    dictionary : dict
        A dictionary (e.g., config file).

    Raises
    ------
    AssertionError
        If dictionary is not a dict or if any value is empty or blank.
    """
    # PRECONDITIONS
    if not isinstance(dictionary, dict):
        raise TypeError(f"dictionary must be a dict, got {type(dictionary)}")

    # MAIN FUNCTION
    for v in dictionary.items():
        if type(v) is str:
            assert v, f'There is an empty key (e.g., ""): {v, dictionary.items()}'
            assert (
                v.strip()
            ), f'There is a blank key (e.g., space, " "): {v, dictionary.items()}'


def normalize_url(host: str, port: int, scheme: str = "http") -> str:
    """
    Normalize the given URL, ensuring it starts with the specified scheme.

    Parameters
    ----------
    host : str
        The host to be normalized.
    port : int
        The port number.
    scheme : str, optional
        The URL scheme (default is "http").

    Returns
    -------
    str
        The normalized URL.

    Raises
    ------
    TypeError
        If host, port, or scheme are not of the correct type, or if URL cannot be normalized.

    Examples
    --------
    >>> normalize_url("localhost", 7474)
    'http://localhost:7474'
    >>> normalize_url("example.com", 80, "bolt")
    'bolt://example.com:80'
    """
    ## PRECONDITIONS
    if not isinstance(host, str):
        raise TypeError(f"host should be a str e.g., 'localhost': {type(host)}")
    if not isinstance(port, int):
        raise TypeError(f"port must be int e.g., '7474': {type(port)}")
    if not isinstance(scheme, str):
        raise TypeError(f"scheme must be str: {type(scheme)}")

    ## MAIN FUNCTION
    if not urlparse(host).netloc:
        host = urlunsplit([scheme, host, "", "", ""])

    # Remove any trailing slashes
    url = host.rstrip("/")

    # Add the port
    url = f"{url}:{str(port)}"

    ## POSTCOND CHECKS
    if not urlparse(url).netloc:
        raise TypeError(f"Unable to normalize url: {url}")

    return url


def get_args(prog_name: str, others: dict = None):
    """
    Initiate argparse.ArgumentParser() and add common arguments.

    Parameters
    ----------
    prog_name : str
        The name of the program.
    others : dict, optional
        Additional keyword arguments for ArgumentParser, by default {}.

    Returns
    -------
    argparse.Namespace
        Parsed command-line arguments.

    Raises
    ------
    TypeError
        If prog_name is not a string or others is not a dict.
    """
    ### PRECONDITIONS
    if not isinstance(prog_name, str):
        raise TypeError(f"prog_name should be a string: {type(prog_name)}")
    if others is None:
        others = {}
    elif not isinstance(others, dict):
        raise TypeError(f"other kwargs must be a dict: {type(others)}")

    ## MAIN FUNCTION
    # init
    parser = argparse.ArgumentParser(prog=prog_name, **others)
    # config file path
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        default="demo/config.yaml",
        help="provide path to config yaml file",
    )
    args = parser.parse_args()
    return args


def get_basename(fname: None | str = None) -> str:
    """
    Get the basename of a given filename, without file extension.

    If no filename is given, returns the basename of the current script.

    Parameters
    ----------
    fname : str or None, optional
        The filename to get basename of, or None (default is None).

    Returns
    -------
    str
        Basename of the given filepath or the current file the function is executed in.
    """
    if fname is not None:
        # PRECONDITION
        assert_path(fname)
        # MAIN FUNCTIONS
        return os.path.splitext(os.path.basename(fname))[0]
    else:
        return os.path.splitext(os.path.basename(sys.argv[0]))[0]


def get_time(incl_time: bool = True, incl_timezone: bool = True) -> str:
    """
    Get current date, time (optional), and timezone (optional) for file naming.

    Parameters
    ----------
    incl_time : bool, optional
        Whether to include timestamp in the string (default is True).
    incl_timezone : bool, optional
        Whether to include the timezone in the string (default is True).

    Returns
    -------
    str
        String including date, timestamp and/or timezone, e.g. 'yyyyMMdd_hhmm_timezone'.

    Raises
    ------
    TypeError
        If incl_time or incl_timezone are not bool.
    AssertionError
        If the output format is not as expected.
    """
    # PRECONDITIONALS
    if not isinstance(incl_time, bool):
        raise TypeError("incl_time must be True or False")
    if not isinstance(incl_timezone, bool):
        raise TypeError("incl_timezone must be True or False")

    # MAIN FUNCTION
    # getting current time and timezone
    the_time = datetime.now()
    timezone = datetime.now().astimezone().tzname()
    # convert date parts to string
    y = str(the_time.year)
    M = str(the_time.month)
    d = str(the_time.day)
    h = str(the_time.hour)
    m = str(the_time.minute)
    s = str(the_time.second)
    # putting date parts into one string
    if incl_time and incl_timezone:
        fname = "_".join([y + M + d, h + m + s, timezone])
    elif incl_time:
        fname = "_".join([y + M + d, h + m + s])
    elif incl_timezone:
        fname = "_".join([y + M + d, timezone])
    else:
        fname = y + M + d

    # POSTCONDITIONALS
    parts = fname.split("_")
    if incl_time and incl_timezone:
        assert len(parts) == 3, f"time and/or timezone inclusion issue: {fname}"
    elif incl_time or incl_timezone:
        assert len(parts) == 2, f"time/timezone inclusion issue: {fname}"
    else:
        assert len(parts) == 1, f"time/timezone inclusion issue: {fname}"

    return fname


def generate_log_filename(folder: str = "logs", suffix: str = "") -> str:
    """
    Create a log file name and path.

    Parameters
    ----------
    folder : str, optional
        Name of the folder to put the log file in (default is "logs").
    suffix : str, optional
        Additional string to add to the log file name (default is "").

    Returns
    -------
    str
        The file path to the log file.
    """
    # PRECONDITIONS
    create_folder(folder)

    # MAIN FUNCTION
    log_filename = get_time(incl_timezone=False) + "_" + suffix + ".log"
    log_filepath = os.path.join(folder, log_filename)

    return log_filepath


def init_log(filename: str, display: bool = False, logger_id: str | None = None):
    """
    Configure a custom Python logger with file and optional stdout handlers.

    Parameters
    ----------
    filename : str
        Filepath to log record file.
    display : bool, optional
        Whether to print the logs to standard output (default is False).
    logger_id : str or None, optional
        An optional identifier for the logger. If None, defaults to 'root'.

    Returns
    -------
    logging.Logger
        Configured logger object.

    Raises
    ------
    TypeError
        If filename is not a string or logger_id is not a string or None.
    """
    # PRECONDITIONS
    if not isinstance(filename, str):
        raise TypeError(f"filename must be a string: {filename}")
    if not (isinstance(logger_id, str) or logger_id is None):
        raise TypeError("logger_id must be a string or None")

    # MAIN FUNCTION
    # init handlers
    file_handler = logging.FileHandler(filename=filename)
    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    if display:
        handlers = [file_handler, stdout_handler]
    else:
        handlers = [file_handler]

    # logger configuration
    logging.basicConfig(
        # level=logging.DEBUG,
        format="[%(asctime)s] %(name)s: %(levelname)s - %(message)s",
        handlers=handlers,
    )
    logging.getLogger("matplotlib.font_manager").disabled = True

    # instantiate the logger
    logger = logging.getLogger(logger_id)
    logger.setLevel(logging.DEBUG)

    return logger


def get_logger():
    """
    Initialize and return a logger with a log file named after the current script.

    Returns
    -------
    logging.Logger
        Configured logger object.
    """
    # get log suffix, which will be the current script's base file name
    log_suffix = get_basename()
    # generate log file name
    log_file = generate_log_filename(suffix=log_suffix)
    # init logger
    logger = init_log(log_file, display=True)
    # log it
    logger.info(f"Path to log file: {log_file}")

    return logger


# FUNCTIONS FOR CONFIG
# def config_loader(filepath: str) -> dict:
#     """
#     Load a YAML config file as a dictionary.

#     Parameters
#     ----------
#     filepath : str
#         Path to the config file.

#     Returns
#     -------
#     dict
#         Configuration parameters as a dictionary.
#     """
#     # PRECONDITIONS
#     assert_path(filepath)

#     # MAIN FUNCTION
#     with open(filepath, "r") as f:
#         contents = yaml.safe_load(f)

#     # POSTCONDITIONS
#     assert isinstance(contents, dict), "content not returned as a dict"

#     return contents



import pytest
from abaco import utils


def test_get_basename():
    assert utils.get_basename("./tests/test_utils.py") == "test_utils"


@pytest.mark.xfail(raises=FileNotFoundError)
def test_assert_path_nonexistent():
    utils.assert_path("this-is-obvi-a-nonexistent/path/blahdeblah")


@pytest.mark.xfail(raises=TypeError)
def test_assert_path_non_string():
    utils.assert_path(123)


@pytest.mark.xfail(reason="the output of get_time() is dynamic")
def test_get_time():
    assert utils.get_time() == "20231019_101758_CEST"



import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from skbio.stats.ordination import pcoa
from scipy.spatial.distance import pdist, squareform
import plotly.graph_objects as go
from plotly.validator_cache import ValidatorCache
from plotly.subplots import make_subplots
from clustergrammer2 import Network, CGM2
from abaco.dataloader import DataTransform


def plotPCoA(
    data,
    method="aitchison",
    sample_label="sample",
    batch_label="batch",
    experiment_label="tissue",
    mode="base",
):
    """
    Plot Principal Coordinates Analysis (PCoA) for batch effect visualization.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    method : str, optional
        Distance metric to use ('aitchison' or 'bray-curtis'), by default 'aitchison'.
    sample_label : str, optional
        Column name for sample identifiers, by default 'sample'.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.
    experiment_label : str, optional
        Column name for experiment/tissue identifiers, by default 'tissue'.
    mode : str, optional
        Plotting mode ('base' for batch+experiment, 'single' for batch only), by default 'base'.

    Returns
    -------
    None
        Displays a Plotly figure.
    """
    if method == "aitchison":
        # CLR transform
        df = DataTransform(
            data,
            factors=[sample_label, batch_label, experiment_label],
            transformation="CLR",
            count=True,
        )

        # log_transformed = np.log(data.select_dtypes(include = "number") + 1e-9)
        # clr_data = log_transformed - log_transformed.mean(axis=1)[:, None]
        # print(clr_data)

        # Extracting numerical data
        df_otu = df.select_dtypes(include="number")
        # Compute Aitchison distances
        distances = pdist(df_otu, "euclidean")
        distances = squareform(distances)

    elif method == "bray-curtis":
        # Extract numeric data (e.g., OTU count data).
        df_otu = data.select_dtypes(include="number")
        # Convert each sample's counts to relative abundances (row sums are normalized to 1)
        # (Handling potential division by zero)
        row_sums = df_otu.sum(axis=1)
        df_rel = df_otu.div(row_sums.replace(0, np.nan), axis=0).fillna(0)
        # Compute Bray-Curtis distances
        distances = pdist(df_rel, metric="braycurtis")
        distances = squareform(distances)

    else:
        raise (ValueError(f"Method provided not valid: {method}"))

    # PCoA
    pcoa_res = pcoa(distances)
    # Construct DataFrame with principal components and metadata
    df_pcoa = pd.DataFrame(pcoa_res.samples[["PC1", "PC2"]], columns=["PC1", "PC2"])
    df_pcoa.index = (
        data.index
    )  # This assures index are the same and both DataFrames are perfectly aligned
    df_pcoa[[sample_label, batch_label, experiment_label]] = data[
        [sample_label, batch_label, experiment_label]
    ]
    # df_pcoa = pd.concat([data[[sample_label, batch_label, experiment_label]], df_pcoa], axis=1)
    # Extracting available symbols to be used per experiment
    SymbolValidator = ValidatorCache.get_validator("scatter.marker", "symbol")
    raw_symbols = SymbolValidator.values[2::12]

    # Defining a set of colors to be used for batches
    raw_colors = [
        "blue",
        "red",
        "green",
        "orange",
        "purple",
        "cyan",
        "magenta",
        "yellow",
        "black",
        "brown",
        "pink",
        "gray",
        "olive",
        "teal",
        "navy",
        "maroon",
        "gold",
        "lime",
        "indigo",
        "violet",
        "coral",
        "slateblue",
        "aquamarine",
        "crimson",
        "sienna",
        "salmon",
        "turquoise",
        "lavender",
        "chocolate",
        "tomato",
        "plum",
        "peru",
        "khaki",
        "orchid",
        "springgreen",
        "steelblue",
        "seagreen",
        "darkblue",
        "darkred",
        "darkgreen",
        "darkorange",
        "darkviolet",
        "mediumblue",
        "mediumvioletred",
        "mediumseagreen",
        "midnightblue",
        "lightblue",
        "lightgreen",
        "lightcoral",
        "peachpuff",
    ]

    # Adding symbol to corresponding experiment
    df_pcoa["marker"] = None
    for n, exp in enumerate(df_pcoa[experiment_label].unique()):
        df_pcoa.loc[df_pcoa[experiment_label] == exp, "marker"] = raw_symbols[n]

    # Adding color to corresponding batch
    df_pcoa["color"] = None
    for n, batch in enumerate(df_pcoa[batch_label].unique()):
        df_pcoa.loc[df_pcoa[batch_label] == batch, "color"] = raw_colors[n]

    # Creating the plotly figure
    fig = go.Figure()

    if mode == "base":
        # Creating a for loop to alocate PCA data points per batch
        for batch in df_pcoa[batch_label].unique():
            # Creating a for loop to alocate data points per experiment in the current batch
            for exp in df_pcoa[experiment_label].unique():
                # Ploting the points corresponding to the current batch and tissue
                fig.add_trace(
                    go.Scatter(
                        x=df_pcoa[
                            (df_pcoa[batch_label] == batch)
                            & (df_pcoa[experiment_label] == exp)
                        ]["PC1"],
                        y=df_pcoa[
                            (df_pcoa[batch_label] == batch)
                            & (df_pcoa[experiment_label] == exp)
                        ]["PC2"],
                        marker=dict(
                            color=df_pcoa[
                                (df_pcoa[batch_label] == batch)
                                & (df_pcoa[experiment_label] == exp)
                            ]["color"],
                            size=8,
                        ),
                        marker_symbol=df_pcoa[
                            (df_pcoa[batch_label] == batch)
                            & (df_pcoa[experiment_label] == exp)
                        ]["marker"],
                        legendgroup=batch,
                        legendgrouptitle_text="Batch {}".format(batch),
                        name=exp,
                        mode="markers",
                    )
                )

        # fig.update_layout(xaxis_range = [-5, 5],
        #                  yaxis_range = [-5, 5])

        return fig.show()

    elif mode == "single":
        # Creating a for loop to alocate PCA data points per batch
        for batch in df_pcoa[batch_label].unique():
            # Ploting the points corresponding to the current batch and tissue
            fig.add_trace(
                go.Scatter(
                    x=df_pcoa[(df_pcoa[batch_label] == batch)]["PC1"],
                    y=df_pcoa[(df_pcoa[batch_label] == batch)]["PC2"],
                    marker=dict(
                        color=df_pcoa[(df_pcoa[batch_label] == batch)]["color"],
                        size=8,
                    ),
                    legendgroup=batch,
                    legendgrouptitle_text="Batch {}".format(batch),
                    name=batch,
                    mode="markers",
                )
            )

        # fig.update_layout(xaxis_range = [-5, 5],
        #                  yaxis_range = [-5, 5])

        return fig.show()


def plotPCA(
    data, sample_label="sample", batch_label="batch", experiment_label="tissue"
):
    """
    Plot Principal Component Analysis (PCA) for batch effect visualization.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str, optional
        Column name for sample identifiers, by default 'sample'.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.
    experiment_label : str, optional
        Column name for experiment/tissue identifiers, by default 'tissue'.

    Returns
    -------
    None
        Displays a Plotly figure.
    """
    # Realize the PCA
    pca = PCA(n_components=2)
    principal_components = pca.fit_transform(data.select_dtypes(include="number"))
    df_pca = pd.DataFrame(data=principal_components, columns=["PC1", "PC2"])

    df_pca.index = (
        data.index
    )  # This assures index are the same and both DataFrames are perfectly aligned

    # Add the labels from batch and other important information
    df_pca[[sample_label, batch_label, experiment_label]] = data[
        [sample_label, batch_label, experiment_label]
    ]

    # Extracting available symbols to be used per experiment
    SymbolValidator = ValidatorCache.get_validator("scatter.marker", "symbol")
    raw_symbols = SymbolValidator.values[2::12]

    # Defining a set of colors to be used for batches
    raw_colors = [
        "blue",
        "red",
        "green",
        "orange",
        "purple",
        "cyan",
        "magenta",
        "yellow",
        "black",
        "brown",
        "pink",
        "gray",
        "olive",
        "teal",
        "navy",
        "maroon",
        "gold",
        "lime",
        "indigo",
        "violet",
        "coral",
        "slateblue",
        "aquamarine",
        "crimson",
        "sienna",
        "salmon",
        "turquoise",
        "lavender",
        "chocolate",
        "tomato",
        "plum",
        "peru",
        "khaki",
        "orchid",
        "springgreen",
        "steelblue",
        "seagreen",
        "darkblue",
        "darkred",
        "darkgreen",
        "darkorange",
        "darkviolet",
        "mediumblue",
        "mediumvioletred",
        "mediumseagreen",
        "midnightblue",
        "lightblue",
        "lightgreen",
        "lightcoral",
        "peachpuff",
    ]

    # Adding symbol to corresponding experiment
    df_pca["marker"] = None
    for n, exp in enumerate(df_pca[experiment_label].unique()):
        df_pca.loc[df_pca[experiment_label] == exp, "marker"] = raw_symbols[n]

    # Adding color to corresponding batch
    df_pca["color"] = None
    for n, batch in enumerate(df_pca[batch_label].unique()):
        df_pca.loc[df_pca[batch_label] == batch, "color"] = raw_colors[n]

    # Creating the plotly figure
    fig = go.Figure()

    # Creating a for loop to alocate PCA data points per batch
    for batch in df_pca[batch_label].unique():
        # Creating a for loop to alocate data points per experiment in the current batch
        for exp in df_pca[experiment_label].unique():
            # Ploting the points corresponding to the current batch and tissue
            fig.add_trace(
                go.Scatter(
                    x=df_pca[
                        (df_pca[batch_label] == batch)
                        & (df_pca[experiment_label] == exp)
                    ]["PC1"],
                    y=df_pca[
                        (df_pca[batch_label] == batch)
                        & (df_pca[experiment_label] == exp)
                    ]["PC2"],
                    marker=dict(
                        color=df_pca[
                            (df_pca[batch_label] == batch)
                            & (df_pca[experiment_label] == exp)
                        ]["color"],
                        size=8,
                    ),
                    marker_symbol=df_pca[
                        (df_pca[batch_label] == batch)
                        & (df_pca[experiment_label] == exp)
                    ]["marker"],
                    legendgroup=batch,
                    legendgrouptitle_text="Batch {}".format(batch),
                    name=exp,
                    mode="markers",
                )
            )

    # fig.update_layout(xaxis_range = [-5, 5],
    #                  yaxis_range = [-5, 5])

    fig.update_layout(legend=dict(font=dict(size=8), itemwidth=30))

    return fig.show()


def plotOTUBox(data, batch_label="batch"):
    """
    Plot boxplots of OTU abundances grouped by batch.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.

    Returns
    -------
    None
        Displays a Plotly figure with dropdown to select OTUs.
    """
    # Extract OTUs columns names
    otu_cols = [col for col in data.columns if col.startswith("OTU")]
    batch_labels = data[batch_label].unique()
    batch_len = len(batch_labels)

    # Converting DataFrame from wide to long
    df_long = pd.melt(
        data,
        id_vars=[batch_label],
        value_vars=otu_cols,
        var_name="OTU",
        value_name="value",
    )

    # Defining a set of colors to be used for batches
    raw_colors = ["blue", "red", "green", "orange", "purple"]

    # Adding color to corresponding batch
    batch_colors = []
    for i in range(batch_len):
        batch_colors.append(raw_colors[i])

    fig = go.Figure()

    # Add traces for each OTU
    for otu in otu_cols:
        for i, batch in enumerate(batch_labels):
            fig.add_trace(
                go.Box(
                    x=df_long[
                        (df_long["OTU"] == otu) & (df_long[batch_label] == batch)
                    ][batch_label],
                    y=df_long[
                        (df_long["OTU"] == otu) & (df_long[batch_label] == batch)
                    ]["value"],
                    marker=dict(
                        color=batch_colors[i]
                    ),  # Apply color to the batch boxplot
                    name=f"Batch {batch}, {otu}",  # Label each trace by the OTU
                    visible=False,  # Set initially to invisible
                )
            )

    # First OTU visible by default
    for i in range(batch_len):
        fig.data[i].visible = True

    # Add dropdown to select which OTU to display
    fig.update_layout(
        xaxis_title="Batch",
        updatemenus=[
            dict(
                buttons=[
                    *[
                        dict(
                            args=[
                                {
                                    "visible": [
                                        (i >= batch_len * idx)
                                        & (i <= batch_len * idx + (batch_len - 1))
                                        for i in range(len(fig.data))
                                    ]
                                }
                            ],  # Toggle visibility
                            label=otu,
                            method="update",
                        )
                        for idx, otu in enumerate(otu_cols)
                    ]
                ],
                direction="down",
                showactive=True,
                xanchor="left",
                y=1.15,
                yanchor="top",
            )
        ],
    )

    return fig.show()


def plotRLE(
    data, sample_label="sample", batch_label="batch", experiment_label="tissue"
):
    """
    Plot Relative Log Expression (RLE) boxplots for each experiment and batch.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str, optional
        Column name for sample identifiers, by default 'sample'.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.
    experiment_label : str, optional
        Column name for experiment/tissue identifiers, by default 'tissue'.

    Returns
    -------
    None
        Displays a Plotly figure with dropdown to select experiments.
    """
    # Extract OTUs column names
    otu_cols = [col for col in data.columns if col.startswith("OTU")]

    # Converting DataFrame from wide to long
    df_long = pd.melt(
        data,
        id_vars=[sample_label, batch_label, experiment_label],
        value_vars=otu_cols,
        var_name="OTU",
        value_name="value",
    )

    # Calculating the medians of each OTU within each experiment
    df_long["medians"] = None
    for OTU in df_long["OTU"].unique():
        for exp in df_long[experiment_label].unique():
            med = np.median(
                df_long[(df_long["OTU"] == OTU) & (df_long[experiment_label] == exp)][
                    "value"
                ]
            )
            df_long.loc[
                (df_long["OTU"] == OTU) & (df_long[experiment_label] == exp), "medians"
            ] = med

    # Incorporating the difference between OTU value in each sample and the median across all samples from the same tissue
    df_long["RLE"] = df_long["value"] - df_long["medians"]

    # Defining a set of colors to be used for batches
    raw_colors = ["blue", "red", "green", "orange", "purple"]

    # Adding color to corresponding batch
    df_long["color"] = None
    for n, batch in enumerate(df_long[batch_label].unique()):
        df_long.loc[df_long[batch_label] == batch, "color"] = raw_colors[n]

    # Generate RLE plots for each experiment
    fig = go.Figure()

    # Add traces for each experiment
    for exp in df_long[experiment_label].unique():
        # Add traces for each batch
        for batch in df_long[batch_label].unique():
            fig.add_trace(
                go.Box(
                    x=df_long[
                        (df_long[experiment_label] == exp)
                        & (df_long[batch_label] == batch)
                    ][sample_label],
                    y=df_long[
                        (df_long[experiment_label] == exp)
                        & (df_long[batch_label] == batch)
                    ]["RLE"],
                    marker_color=df_long[
                        (df_long[experiment_label] == exp)
                        & (df_long[batch_label] == batch)
                    ]["color"].iloc[0],
                    name="Batch {}".format(batch),  # Label each trace by the batch
                    visible=False,  # Set initially to invisible
                )
            )

    # First experiment's traces visible by default
    for i in range(len(df_long[batch_label].unique())):
        fig.data[i].visible = True

    # Add dropdown to select which experiment to display
    fig.update_layout(
        updatemenus=[
            dict(
                buttons=[
                    *[
                        dict(
                            args=[
                                {
                                    "visible": [
                                        i // len(df_long[batch_label].unique()) == idx
                                        for i in range(len(fig.data))
                                    ]
                                }
                            ],  # Toggle visibility
                            label=exp,
                            method="update",
                        )
                        for idx, exp in enumerate(df_long[experiment_label].unique())
                    ]
                ],
                direction="down",
                showactive=True,
                xanchor="left",
                y=1.15,
                yanchor="top",
            )
        ]
    )

    # Add horizontal dashed red line at y = 0 as a reference point
    fig.add_shape(
        type="line",
        x0=0,
        x1=1,  # Extend the line across the x-axis
        y0=0,
        y1=0,  # Line positioned at y = 0
        xref="paper",
        yref="y",  # "paper" allows the line to span the entire plot width
        line=dict(color="red", width=2, dash="dash"),  # Dashed red line
    )

    return fig.show()


def plotClusterHeatMap(
    data, batch_label="batch", experiment_label="tissue", sample_label="sample"
):
    """
    Plot a clustered heatmap of scaled OTU data with batch and experiment metadata.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.
    experiment_label : str, optional
        Column name for experiment/tissue identifiers, by default 'tissue'.
    sample_label : str, optional
        Column name for sample identifiers, by default 'sample'.

    Returns
    -------
    clustergrammer2.CGM2Widget
        Clustergrammer2 widget displaying the clustered heatmap.
    """
    # Extracts numerical and categorical data of interest
    data_num = data.select_dtypes(include="number")
    data_num.index = [str(i) for i in data[sample_label]]
    data_cat = data[[batch_label, experiment_label]]
    data_cat.index = [str(i) for i in data[sample_label]]

    # First scaling process - Ensures every observation is scaled according to OTUs
    scaler = StandardScaler()
    scaled_data = scaler.fit_transform(data_num)
    scaled_data = pd.DataFrame(
        scaled_data, columns=data_num.columns, index=data_num.index
    )

    # Second scaling process - Ensures every observation is scaled according to sample
    scaler = StandardScaler()
    scaled_data = scaler.fit_transform(scaled_data.T)
    scaled_data = pd.DataFrame(
        scaled_data,
        index=[str(i) for i in data_num.columns],
        columns=[str(i) for i in data_num.index],
    )

    # Create Clustergrammer2 plot
    n2 = Network(CGM2)
    n2.load_df(scaled_data, meta_col=data_cat)
    n2.cluster()
    return n2.widget()


def plot_LISI_perplexity(
    df_c,
    df_i,
    n_samples: int,
    x_col: str = "perplexity",
    y_col_c: str = "cLISI",
    y_col_i: str = "iLISI",
    title_c: str = "Biological conservation (cLISI)",
    title_i: str = "Batch mixing (iLISI)",
):
    """
    Plot cLISI and iLISI scores as a function of perplexity.

    Parameters
    ----------
    df_c : pandas.DataFrame
        DataFrame containing cLISI scores and perplexity values.
    df_i : pandas.DataFrame
        DataFrame containing iLISI scores and perplexity values.
    n_samples : int
        Number of samples in the dataset.
    x_col : str, optional
        Column name for perplexity values, by default 'perplexity'.
    y_col_c : str, optional
        Column name for cLISI scores, by default 'cLISI'.
    y_col_i : str, optional
        Column name for iLISI scores, by default 'iLISI'.
    title_c : str, optional
        Title for the cLISI subplot, by default 'Biological conservation (cLISI)'.
    title_i : str, optional
        Title for the iLISI subplot, by default 'Batch mixing (iLISI)'.

    Returns
    -------
    plotly.graph_objs._figure.Figure
        Plotly figure with cLISI and iLISI subplots.
    """
    ideal_k = max(int(np.sqrt(n_samples)), 1)

    # x-axis limits
    xmin = min(df_c[x_col].min(), df_i[x_col].min()) - 0.5
    xmax = max(df_c[x_col].max(), df_i[x_col].max()) + 0.5

    # y-intersection
    y_c = np.interp(ideal_k, df_c[x_col].values, df_c[y_col_c].values)
    y_i = np.interp(ideal_k, df_i[x_col].values, df_i[y_col_i].values)

    # Create 1×2 subplots
    fig = make_subplots(
        rows=1,
        cols=2,
        shared_yaxes=False,
        shared_xaxes=True,
        subplot_titles=(title_c, title_i),
    )

    # cLISI trace
    fig.add_trace(
        go.Scatter(
            x=df_c[x_col],
            y=df_c[y_col_c],
            mode="lines+markers",
            name="cLISI",
            line=dict(color="green"),
            marker=dict(color="green"),
        ),
        row=1,
        col=1,
    )
    # ideal point in cLISI
    fig.add_shape(
        dict(
            type="line",
            x0=ideal_k,
            y0=0,
            x1=ideal_k,
            y1=y_c,
            line=dict(color="red", dash="dash"),
        ),
        row=1,
        col=1,
    )
    fig.add_shape(
        dict(
            type="line",
            x0=xmin,
            y0=y_c,
            x1=ideal_k,
            y1=y_c,
            line=dict(color="red", dash="dash"),
        ),
        row=1,
        col=1,
    )
    fig.add_annotation(
        dict(
            x=ideal_k,
            y=y_c,
            xref="x1",
            yref="y1",
            text=f"{y_c:.2f}",
            showarrow=True,
            arrowhead=1,
            ax=30,
            ay=-30,
            font=dict(color="red"),
        )
    )

    # iLISI trace
    fig.add_trace(
        go.Scatter(
            x=df_i[x_col],
            y=df_i[y_col_i],
            mode="lines+markers",
            name="iLISI",
            line=dict(color="blue"),
            marker=dict(color="blue"),
        ),
        row=1,
        col=2,
    )
    # Ideal line in iLISI
    fig.add_shape(
        dict(
            type="line",
            x0=ideal_k,
            y0=0,
            x1=ideal_k,
            y1=y_i,
            line=dict(color="red", dash="dash"),
        ),
        row=1,
        col=2,
    )
    fig.add_shape(
        dict(
            type="line",
            x0=xmin,
            y0=y_i,
            x1=ideal_k,
            y1=y_i,
            line=dict(color="red", dash="dash"),
        ),
        row=1,
        col=2,
    )
    fig.add_annotation(
        dict(
            x=ideal_k,
            y=y_i,
            xref="x2",
            yref="y2",
            text=f"{y_i:.2f}",
            showarrow=True,
            arrowhead=1,
            ax=30,
            ay=-30,
            font=dict(color="red"),
        )
    )

    # Update axes
    fig.update_xaxes(title_text="Perplexity (k)", range=[xmin, xmax], row=1, col=1)
    fig.update_xaxes(title_text="Perplexity (k)", range=[xmin, xmax], row=1, col=2)
    fig.update_yaxes(title_text="Normalized cLISI", row=1, col=1)
    fig.update_yaxes(title_text="Normalized iLISI", row=1, col=2)

    fig.update_layout(template="plotly_white", showlegend=False)
    return fig



import pandas as pd
import numpy as np
from scipy.stats import gmean
from sklearn.preprocessing import StandardScaler
import torch
from torch.utils.data import DataLoader, TensorDataset
from abaco.utils import assert_path


def DataPreprocess(
    path: str, factors: list = ["sample", "batch", "tissue"], delimiter: str = ","
):
    """
    Reads a CSV file and preprocesses the data by converting specified columns to categorical type.
    Parameters
    ----------
    path : str
        The path to the CSV file containing the data.
    factors : list, optional
        List of factor columns to convert to categorical type. Default is ["sample", "batch", "tissue"].
    Returns
    -------
    pd.DataFrame
        The preprocessed DataFrame with specified factor columns converted to categorical type.
    """
    # PRECONDITION CHECKS
    assert_path(path)
    if not isinstance(factors, list):
        raise TypeError("Factors should be a list of column names.")

    # MAIN FUNCTION
    # Read the CSV file into a DataFrame
    df = pd.read_csv(path, sep=delimiter)
    # check if factors are in the DataFrame
    for factor in factors:
        if factor not in df.columns:
            raise ValueError(f"Factor '{factor}' not found in the DataFrame columns.")
    # Convert specified columns to categorical type
    df[factors] = df[factors].astype("category")

    return df


def DataTransform(
    data, factors=["sample", "batch", "tissue"], transformation="CLR", count=False
):
    """
    Transforms the data based on the specified transformation method.

    Parameters
    ----------
    data : pd.DataFrame
        The input data containing OTU counts and factors.
    factors : list, optional
        List of factor columns to retain in the transformed data.
    transformation : str, optional
        The transformation method to apply. Options are "CLR", "Sqrt", "ILR", "ALR".
    count : bool, optional
        If True, the data is treated as count data; otherwise, a small offset is added
        to avoid log(0) issues.
    Returns
    -------
    pd.DataFrame
        The transformed data with the specified factors and transformed OTU counts.
    """
    if transformation == "CLR":
        if not count:
            # Select only OTUs columns and adding a small offset
            df_otu = data.select_dtypes(include="number") + 1e-9

        else:
            df_otu = data.select_dtypes(include="number") + 1

        # Apply CLR transformation to numeric columns
        df_clr = np.log(df_otu.div(gmean(df_otu, axis=1), axis=0))

        # Combine CLR-transformed data with non-numeric columns
        df = pd.concat([data[factors], df_clr], axis=1)

    elif transformation == "Sqrt":
        # Select only OTUs columns
        df_otu = data.select_dtypes(include="number")

        # Apply Square-root transformation to numeric columns
        df_sqrt = np.sqrt(df_otu)

        # Standardize the squared-rooted data
        scaler = StandardScaler()
        df_stsqrt = scaler.fit_transform(df_sqrt)

        # Convert back to DataFrame
        df_stsqrt = pd.DataFrame(df_stsqrt, columns=df_sqrt.columns)

        # Combine data with non-numeric columns
        df = pd.concat([data[factors], df_stsqrt], axis=1)

    elif transformation == "ILR":
        print("Not yet developed")

    elif transformation == "ALR":
        print("To be developed")

    else:
        raise (ValueError(f"Not a valid transformation: {transformation}"))

    return df


def DataReverseTransform(
    data,
    original_data,
    factors=["sample", "batch", "tissue"],
    transformation="CLR",
    count=False,
):
    if transformation == "CLR":
        df_otu = data.select_dtypes(include="number")
        if not count:
            df_otu_original = original_data.select_dtypes(include="number") + 1e-9

        else:
            df_otu_original = original_data.select_dtypes(include="number") + 1

        df_inv = round(np.exp(df_otu) * gmean(df_otu_original, axis=1)[:, None], 3)

        df = pd.concat([data[factors], df_inv], axis=1)

    return df


def one_hot_encoding(labels: pd.Series, dtype: torch.dtype = torch.float32) -> tuple:
    """
    Converts a series of labels into a one-hot encoded matrix.

    Parameters
    ----------
    labels : pd.Series
        The input labels to be one-hot encoded.
    dtype : torch.dtype, optional
        The data type of the output tensor. Default is torch.float32.

    Returns
    -------
    tuple
        A tuple containing:
        - torch.Tensor: A one-hot encoded matrix where each row corresponds to a label.
        - list: The categories (unique labels) encoded in matrix

    Example
    -------
    >>> import pandas as pd
    >>> import torch
    >>> labels = pd.Series(['A', 'B', 'A', 'C'])
    >>> one_hot_matrix, categories = one_hot_encoding(labels, dtype=torch.int32)
    >>> print(one_hot_matrix)
    tensor([[1, 0, 0],
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 1]], dtype=torch.int32)
    >>> print(categories)
    ['A', 'B', 'C']
    """
    # Ensure labels are a pandas Series
    if not isinstance(labels, pd.Series):
        raise TypeError("Input labels must be a pandas Series.")

    alphabet = labels.unique()
    label_to_int = {label: i for i, label in enumerate(alphabet)}

    # Initialize the one-hot encoded matrix
    one_hot = np.zeros((len(labels), len(alphabet)), dtype=int)

    # Fill the matrix
    for i, label in enumerate(labels):
        if label in label_to_int:
            one_hot[i, label_to_int[label]] = 1

    return torch.tensor(one_hot, dtype=dtype), alphabet.tolist()


def class_to_int(labels):
    # Dictionary of batch labels
    alphabet = labels.unique()
    label_to_int = {label: i for i, label in enumerate(alphabet)}

    # Initialize the empty array
    classes = np.zeros(len(labels), dtype=int)

    # Fill the matrix
    for i, label in enumerate(labels):
        classes[i] = label_to_int[label]

    return torch.tensor(classes)


def ABaCoDataLoader(
    data,
    device=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
    batch_label="batch",
    exp_label="tissue",
    batch_size=32,
    total_size=1024,
    total_batch=10,
):
    # Convert data to tensor (structure: tensor([otus], [batch]))
    otu_data = data.select_dtypes(include="number")
    otu_tensor = torch.tensor(otu_data.values, dtype=torch.float32)

    # Add zero padding for input
    n, m = otu_tensor.shape
    zero_padding = torch.zeros((n, total_size - m))

    # Extract labels and convert to one hot encoding matrix
    data_batch = data[batch_label]
    data_tissue = data[exp_label]
    ohe_batch, _ = one_hot_encoding(data_batch)
    ohe_tissue, _ = one_hot_encoding(data_tissue)

    # Add zero padding for batch input
    k, j = ohe_batch.shape
    batch_padding = torch.zeros((k, total_batch - j))

    # Send to device
    otu_tensor = otu_tensor.to(device)
    ohe_batch = ohe_batch.to(device)
    ohe_tissue = ohe_tissue.to(device)
    zero_padding = zero_padding.to(device)
    batch_padding = batch_padding.to(device)

    # otu_dataloader = DataLoader(otu_tensor, batch_size = batch_size)
    # batch_dataloader = DataLoader(ohe_batch, batch_size = batch_size)
    # tissue_dataloader = DataLoader(ohe_tissue, batch_size = batch_size)

    # Defining DataLoader for otus + batch information
    otu_tensor_padded = torch.concat((otu_tensor, zero_padding), 1)
    ohe_batch_padded = torch.concat((ohe_batch, batch_padding), 1)

    otu_batch_tensor = torch.concat((otu_tensor_padded, ohe_batch_padded), 1)
    # otu_batch_dataloader = DataLoader(otu_batch_tensor, batch_size = batch_size)

    # Defining DataLoader for otus + tissue information, also including batch as label for discriminator training
    # otu_tissue_tensor = torch.concat((otu_tensor, ohe_tissue), 1)
    # otu_tissue_dataloader = DataLoader(TensorDataset(otu_tissue_tensor, class_to_int(data_batch)), batch_size = batch_size)

    # Defining DataLoader for otus including tissue as label for classifier training
    # otu_tissue_class_dataloader = DataLoader(TensorDataset(otu_tensor, class_to_int(data_tissue)), batch_size = batch_size)

    # Defining DataLoader for otus including + batch information, also including tissue as label for classificator training
    k_features = torch.full((n,), m, device=device)
    abaco_dataloader = DataLoader(
        TensorDataset(
            otu_batch_tensor, class_to_int(data_tissue).to(device), k_features
        ),
        batch_size=batch_size,
    )
    ohe_dataloader = DataLoader(ohe_tissue, batch_size=batch_size)

    return (
        abaco_dataloader,
        ohe_batch,
        ohe_dataloader,
        otu_data,
        data_batch,
        data_tissue,
    )



import pandas as pd
import numpy as np
from combat.pycombat import pycombat
from inmoose.pycombat import pycombat_seq
import statsmodels.api as sm
from sklearn.preprocessing import scale
from sklearn.cross_decomposition import PLSRegression
from sklearn.linear_model import LogisticRegression, QuantileRegressor
from sklearn.base import TransformerMixin, BaseEstimator


# Batch Mean Centering
def correctBMC(data, sample_label, batch_label, exp_label):
    """
    This function, LITERALLY, substracts the mean of each batch (group) from each feature.
    Perform Batch Mean Centering (BMC) correction.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str
        Column name for sample identifiers.
    batch_label : str
        Column name for batch identifiers.
    exp_label : str
        Column name for experiment/tissue identifiers.

    Returns
    -------
    pandas.DataFrame
        DataFrame with sample, experiment, batch, and batch mean centered features.
    """
    features = data.select_dtypes(include="number")
    batch_means = features.groupby(data[batch_label]).transform("mean")
    corrected = features - batch_means
    df_all = pd.concat(
        [data[sample_label], data[exp_label], data[batch_label], corrected], axis=1
    )
    return df_all


# ComBat
def correctCombat(
    data, sample_label="sample", batch_label="batch", experiment_label="tissue"
):
    """
    Perform ComBat batch correction.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str, optional
        Column name for sample identifiers, by default 'sample'.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.
    experiment_label : str, optional
        Column name for experiment/tissue identifiers, by default 'tissue'.

    Returns
    -------
    pandas.DataFrame
        DataFrame with sample, batch, experiment, and ComBat-corrected features.
    """
    num_data = data.select_dtypes(include="number")
    batch_data = [batch for batch in data[batch_label]]
    cov_data = [exp for exp in data[experiment_label]]

    corrected_data = pycombat(num_data.T, batch_data, cov_data)
    data_combat = pd.concat(
        [data[[sample_label, batch_label, experiment_label]], corrected_data.T], axis=1
    )

    return data_combat


# Limma (removeBatchEffect)
def correctLimma_rBE(
    data, sample_label="sample", batch_label="batch", covariates_labels=None
):
    """
    Perform batch correction using Limma's removeBatchEffect approach.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str, optional
        Column name for sample identifiers, by default 'sample'.
    batch_label : str, optional
        Column name for batch identifiers, by default 'batch'.
    covariates_labels : str or list of str, optional
        Additional covariate column(s) to include in the model.

    Returns
    -------
    pandas.DataFrame
        DataFrame with original labels and batch-corrected numeric data.
    """
    # Extract numeric variables from data
    num_data = data.select_dtypes(include="number")

    # Convert batch labels to one-hot encoded DataFrame for regression;
    # drop_first=True to avoid multicollinearity but keep all resulting column names intact.
    batch = pd.get_dummies(data[batch_label], drop_first=True)

    # Combine batch and covariates if provided
    if covariates_labels is not None:
        # Get dummy variables for covariates (if more than one, this returns multiple columns)
        covariates = pd.get_dummies(data[covariates_labels], drop_first=True)
        # Build the full design matrix (batch effects + covariates)
        design_matrix = pd.concat([batch, covariates], axis=1)

        # For the batch-only design matrix, keep the batch part and fill zeros for covariates
        zeros_cov = pd.DataFrame(
            np.zeros_like(covariates),
            columns=covariates.columns,
            index=covariates.index,
        )
        design_matrix_batch = pd.concat([batch, zeros_cov], axis=1)
        # Add constant to the design matrix
        design_matrix_batch = sm.add_constant(design_matrix_batch, has_constant="add")
        design_matrix_batch = design_matrix_batch.astype(float)
    else:
        design_matrix = batch.copy()
        design_matrix_batch = design_matrix.copy()

    # Ensure an intercept (constant) is added to the full design matrix
    design_matrix = sm.add_constant(design_matrix, has_constant="add")
    design_matrix = design_matrix.astype(float)

    # Initialize a DataFrame to store batch-corrected values
    corrected_data = pd.DataFrame(index=num_data.index, columns=num_data.columns)

    # Regress out batch effect for each feature
    for feature in num_data.columns:
        model = sm.OLS(num_data[feature], design_matrix).fit()

        # Predict the portion attributable to batch effects only using the design_matrix_batch
        batch_effect = model.predict(design_matrix_batch)

        # Subtract the estimated batch effect from the original feature values
        corrected_data[feature] = num_data[feature] - batch_effect

    # Prepare final DataFrame with original labels and corrected numeric data.
    # If there are covariates, include them as well.
    columns_to_keep = [sample_label, batch_label]
    if covariates_labels is not None:
        if isinstance(covariates_labels, list):
            columns_to_keep += covariates_labels
        else:
            columns_to_keep.append(covariates_labels)

    # Combine the metadata with the corrected numeric data.
    data_limma = pd.concat(
        [
            data[columns_to_keep].reset_index(drop=True),
            corrected_data.reset_index(drop=True),
        ],
        axis=1,
    )

    return data_limma


# PLSDA-batch
def correctPLSDAbatch(
    df: pd.DataFrame,
    sample_label: str,
    exp_label: str,
    batch_label: str,
    ncomp_trt: int = 1,
    ncomp_batch: int = 1,
):
    """
    Perform PLSDA-batch correction.

    Parameters
    ----------
    df : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str
        Column name for sample identifiers.
    exp_label : str
        Column name for experiment/tissue identifiers.
    batch_label : str
        Column name for batch identifiers.
    ncomp_trt : int, optional
        Number of treatment components, by default 1.
    ncomp_batch : int, optional
        Number of batch components, by default 1.

    Returns
    -------
    pandas.DataFrame
        DataFrame with sample, experiment, batch, and PLSDA-batch corrected features.
    """
    y_sample = df[sample_label]
    y_trt = df[exp_label]
    y_batch = df[batch_label]
    df = df.select_dtypes(include="number")
    X = df.values  # (n, p)
    # Step 1: Encode outcomes
    Y_trt = pd.get_dummies(y_trt).values  # (n, n_trt)
    Y_batch = pd.get_dummies(y_batch).values  # (n, n_batch)

    # Step 2: Fit Partial Least Squares for treatment
    pls_trt = PLSRegression(n_components=ncomp_trt, scale=True)
    pls_trt.fit(X, Y_trt)
    T_trt = pls_trt.x_scores_  # (n, ncomp_trt)
    P_trt = pls_trt.x_loadings_  # (p, ncomp_trt)

    # Step 3: Deflate X by treatment components
    X_res = X - T_trt @ P_trt.T  # (n, p)

    # Step 4: Fit Partial Least Squares for batch on residuals
    pls_batch = PLSRegression(n_components=ncomp_batch, scale=True)
    pls_batch.fit(X_res, Y_batch)
    T_batch = pls_batch.x_scores_  # (n, ncomp_trt)
    P_batch = pls_batch.x_loadings_  # (p, ncomp_trt)

    # Step 5: Substract batch variation from orinial X
    X_nobatch = X - T_batch @ P_batch.T

    # Return as pd.DataFrame
    df_corrected = pd.DataFrame(
        X_nobatch,
        index=df.index,
        columns=df.columns,
    )
    df_all = pd.concat([y_trt, y_batch, y_sample, df_corrected], axis=1)
    return df_all


# PLSDA-batch analogous to R implementation
def deflate_mtx(X, t):
    """
    Deflate matrix X by component t: X - t (t^T t)^{-1} t^T X

    Parameters
    ----------
    X : numpy.ndarray
        Data matrix to be deflated.
    t : numpy.ndarray
        Component vector.

    Returns
    -------
    numpy.ndarray
        Deflated matrix.
    """
    # t: (n_samples,)
    denom = t.T @ t
    if denom == 0:
        return X.copy()
    proj = np.outer(t, (t.T @ X) / denom)
    return X - proj


class PLSDA:
    """
    Partial Least Squares Discriminant Analysis (PLSDA) implementation.

    Parameters
    ----------
    ncomp : int, optional
        Number of components to extract, by default 1.
    keepX : list of int, optional
        Number of variables to keep for each component (for sparsity).
    tol : float, optional
        Convergence tolerance, by default 1e-6.
    max_iter : int, optional
        Maximum number of iterations, by default 500.

    Attributes
    ----------
    t_ : numpy.ndarray
        X scores.
    u_ : numpy.ndarray
        Y scores.
    a_ : numpy.ndarray
        X loadings.
    b_ : numpy.ndarray
        Y loadings.
    iters_ : list
        Number of iterations per component.
    exp_var_ : list
        Explained variance per component.
    """

    def __init__(self, ncomp=1, keepX=None, tol=1e-6, max_iter=500):
        self.ncomp = ncomp
        self.keepX = keepX or []
        self.t_ = None
        self.u_ = None
        self.a_ = None
        self.b_ = None
        self.iters_ = []
        self.exp_var_ = None
        self.max_iter = max_iter
        self.tol = tol

    def fit(self, X, Y):
        # X, Y are centered and scaled numpy arrays
        n, p = X.shape
        _, q = Y.shape
        keepX = self.keepX if len(self.keepX) == self.ncomp else [p] * self.ncomp
        T = np.zeros((n, self.ncomp))
        U = np.zeros((n, self.ncomp))
        A = np.zeros((p, self.ncomp))
        B = np.zeros((q, self.ncomp))
        X_temp = X.copy()
        Y_temp = Y.copy()
        for h in range(self.ncomp):
            # initial SVD-based loadings
            M = X_temp.T @ Y_temp
            u, s, vh = np.linalg.svd(M, full_matrices=False)
            a = u[:, 0]
            b = vh.T[:, 0]
            # normalize loadings
            a = a / np.linalg.norm(a)
            b = b / np.linalg.norm(b)
            t = X_temp @ a
            u_comp = Y_temp @ b
            # iterative NIPALS
            it = 0
            while it < self.max_iter:
                it += 1
                # update a
                a_new = X_temp.T @ u_comp
                if keepX[h] < p:
                    # sparsity: zero smallest abs entries
                    abs_a = np.abs(a_new)
                    thresh = np.sort(abs_a)[-keepX[h]] if keepX[h] > 0 else np.inf
                    a_new[abs_a < thresh] = 0
                a_new = a_new / np.linalg.norm(a_new)
                t = X_temp @ a_new
                b_new = Y_temp.T @ t
                b_new = b_new / np.linalg.norm(b_new)
                u_comp = Y_temp @ b_new
                if np.linalg.norm(a_new - a) < self.tol:
                    break
                a, b = a_new, b_new
            # store component
            T[:, h] = t
            U[:, h] = u_comp
            A[:, h] = a_new
            B[:, h] = b_new
            self.iters_.append(it)
            # deflate
            X_temp = deflate_mtx(X_temp, t)
            Y_temp = deflate_mtx(Y_temp, u_comp)
        self.t_ = T
        self.u_ = U
        self.a_ = A
        self.b_ = B
        # explained variance on X
        tot_var = np.sum(X**2)
        var_expl = [
            np.sum((T[:, i : i + 1] @ A[:, i : i + 1].T) ** 2) / tot_var
            for i in range(self.ncomp)
        ]
        self.exp_var_ = var_expl
        return self


def correctPLSDAbatch_R(
    df,
    sample_label,
    exp_label,
    batch_label,
    ncomp_trt=1,
    ncomp_bat=1,
    keepX_trt=None,
    keepX_bat=None,
    tol=1e-6,
    max_iter=500,
    near_zero_var=True,
    balance=True,
):
    """
    Python adaptation of PLSDA_batch from R. Returns corrected DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        Input data containing OTU counts and metadata.
    sample_label : str
        Column name for sample identifiers.
    exp_label : str
        Column name for experiment/tissue identifiers.
    batch_label : str
        Column name for batch identifiers.
    ncomp_trt : int, optional
        Number of treatment components, by default 1.
    ncomp_bat : int, optional
        Number of batch components, by default 1.
    keepX_trt : list of int, optional
        Number of variables to keep for each treatment component.
    keepX_bat : list of int, optional
        Number of variables to keep for each batch component.
    tol : float, optional
        Convergence tolerance, by default 1e-6.
    max_iter : int, optional
        Maximum number of iterations, by default 500.
    near_zero_var : bool, optional
        Whether to filter near-zero variance features, by default True.
    balance : bool, optional
        Whether to balance design, by default True.

    Returns
    -------
    pandas.DataFrame
        DataFrame with sample, experiment, batch, and corrected features.
    """
    y_sample = df[sample_label]
    y_trt = df[exp_label]
    y_batch = df[batch_label]
    df = df.select_dtypes(include="number")
    X = df.values.copy()
    n, p = X.shape
    # near zero variance filtering
    if near_zero_var:
        var = X.var(axis=0)
        keep = var > 1e-8
        X = X[:, keep]
        cols = df.columns[keep]
    else:
        cols = df.columns
    # encode Y
    Y_trt = pd.get_dummies(y_trt).values
    Y_bat = pd.get_dummies(y_batch).values
    # weighting
    weight = np.ones(n)
    if not balance:
        # implement weighted design (omitted for brevity)
        pass
    # scale
    Xs = scale(X, with_mean=True, with_std=True)
    Ys_trt = scale(weight[:, None] * Y_trt)
    Ys_bat = scale(Y_bat)
    # stage1: treatment
    pls_trt = PLSDA(
        ncomp=ncomp_trt, keepX=keepX_trt or [p] * ncomp_trt, tol=tol, max_iter=max_iter
    )
    pls_trt.fit(Xs, Ys_trt)
    X_notrt = (
        deflate_mtx(Xs, pls_trt.t_[:, 0])
        if ncomp_trt == 1
        else Xs - pls_trt.t_ @ pls_trt.a_.T
    )
    # stage2: batch
    pls_bat = PLSDA(
        ncomp=ncomp_bat, keepX=keepX_bat or [p] * ncomp_bat, tol=tol, max_iter=max_iter
    )
    pls_bat.fit(X_notrt, Ys_bat)
    # deflate all batch components from Xs
    X_temp = Xs.copy()
    for h in range(ncomp_bat):
        X_temp = deflate_mtx(X_temp, pls_bat.t_[:, h])
    # back-transform
    # unscale
    X_nobat = X_temp * np.std(X, axis=0) + np.mean(X, axis=0)
    df_corr = pd.DataFrame(X_nobat, index=df.index, columns=cols)
    df_all = pd.concat([y_sample, y_trt, y_batch, df_corr], axis=1)
    return df_all


# ConQuR - analog to R function and PyPI implementation


class ConQur(TransformerMixin, BaseEstimator):
    """
    Conditional Quantile Regression (ConQuR) batch correction transformer.

    Parameters
    ----------
    batch_cols : list of str
        List of batch column names.
    covariate_cols : list of str
        List of covariate column names.
    reference_batch : dict
        Dictionary specifying reference batch values for each batch column.
    quantiles : tuple of float, optional
        Quantiles to use for quantile regression, by default (0.05, 0.5, 0.95).
    logistic_kwargs : dict, optional
        Keyword arguments for LogisticRegression.
    quantile_kwargs : dict, optional
        Keyword arguments for QuantileRegressor.

    Attributes
    ----------
    _logit_models : dict
        Fitted logistic regression models for zero-mass.
    _quantile_models : dict
        Fitted quantile regression models for nonzero values.
    _col_order : list
        Order of columns used in the model.
    _feature_cols : list
        List of feature columns.
    """

    def __init__(
        self,
        batch_cols,
        covariate_cols,
        reference_batch,  # e.g. {'batch': 0}
        quantiles=(0.05, 0.5, 0.95),
        logistic_kwargs=None,
        quantile_kwargs=None,
    ):
        self.batch_cols = batch_cols
        self.covariate_cols = covariate_cols
        self.reference_batch = reference_batch
        self.quantiles = np.array(quantiles)
        self.logistic_kwargs = logistic_kwargs or {}
        self.quantile_kwargs = quantile_kwargs or {}
        self._logit_models = {}
        self._quantile_models = {}
        self._col_order = []
        self._feature_cols = []

    def fit(self, df, y=None):
        df = df.copy()
        # encode batch/covariates
        for col in self.batch_cols + self.covariate_cols:
            if not pd.api.types.is_numeric_dtype(df[col]):
                df[col] = pd.Categorical(df[col]).codes

        # identify features
        reserved = set(self.batch_cols + self.covariate_cols)
        numeric = df.select_dtypes(include="number").columns.tolist()
        self._feature_cols = [c for c in numeric if c not in reserved]
        # define column order in arrays
        self._col_order = self.batch_cols + self.covariate_cols + self._feature_cols

        X_full = df[self._col_order].values.astype(float)
        # build reference array
        X_ref = X_full.copy()
        for col, ref in self.reference_batch.items():
            idx = self._col_order.index(col)
            X_ref[:, idx] = ref

        # design matrices
        bc = [self._col_order.index(c) for c in self.batch_cols]
        cc = [self._col_order.index(c) for c in self.covariate_cols]
        design_idx = bc + cc
        Xd = X_full[:, design_idx]
        Xd_ref = X_ref[:, design_idx]

        n, p = X_full.shape
        feat_idx = list(range(len(design_idx), p))

        # fit per‐feature models
        for f in feat_idx:
            yv = X_full[:, f]
            ybin = (yv != 0).astype(int)

            # logistic
            if ybin.sum() == 0:
                self._logit_models[f] = ("all_zero", None)
            elif ybin.sum() == n:
                self._logit_models[f] = ("all_one", None)
            else:
                lr = LogisticRegression(**self.logistic_kwargs)
                lr.fit(Xd, ybin)
                self._logit_models[f] = ("model", lr)

            # quantile models on nonzero
            mask = yv != 0
            Xnz = Xd[mask]
            ynz = yv[mask]
            qmods = {
                q: QuantileRegressor(quantile=q, **self.quantile_kwargs).fit(Xnz, ynz)
                for q in self.quantiles
            }
            self._quantile_models[f] = qmods

        return self

    def transform(self, df):
        df_base = df
        df = df.copy()
        # encode again
        for col in self.batch_cols + self.covariate_cols:
            if not pd.api.types.is_numeric_dtype(df[col]):
                df[col] = pd.Categorical(df[col]).codes

        X_full = df[self._col_order].values.astype(float)
        X_ref = X_full.copy()
        for col, ref in self.reference_batch.items():
            idx = self._col_order.index(col)
            X_ref[:, idx] = ref

        bc = [self._col_order.index(c) for c in self.batch_cols]
        cc = [self._col_order.index(c) for c in self.covariate_cols]
        design_idx = bc + cc
        Xd = X_full[:, design_idx]
        Xd_ref = X_ref[:, design_idx]

        n, p = X_full.shape
        feat_idx = list(range(len(design_idx), p))
        Xcorr = X_full.copy()

        # apply mapping
        for f in feat_idx:
            yv = X_full[:, f]
            tag, lm = self._logit_models[f]

            # zero‐mass probabilities
            if tag == "all_zero":
                p0 = np.ones(n)
                p0_ref = np.ones(n)
            elif tag == "all_one":
                p0 = np.zeros(n)
                p0_ref = np.zeros(n)
            else:
                p0 = lm.predict_proba(Xd)[:, 0]
                p0_ref = lm.predict_proba(Xd_ref)[:, 0]

            # nonzero quantile predictions
            qmods = self._quantile_models[f]
            Q = np.vstack([qmods[q].predict(Xd) for q in self.quantiles]).T
            Q_ref = np.vstack([qmods[q].predict(Xd_ref) for q in self.quantiles]).T

            # for each sample compute corrected value
            new = np.zeros_like(yv, dtype=float)
            for i in range(n):
                if yv[i] == 0:
                    # keep zero if p0_ref high, else use smallest nonzero quantile
                    new[i] = 0.0 if p0_ref[i] >= 0.5 else Q_ref[i, 0]
                else:
                    # find conditional quantile rank q
                    # solve F = p0 + (1−p0)*q  ⇒  q = (F−p0)/(1−p0)
                    # approximate F by finding which quantile bin y falls into:
                    diffs = np.abs(Q[i] - yv[i])
                    bin_idx = np.argmin(diffs)
                    qbin = self.quantiles[bin_idx]
                    # overall CDF position
                    F = p0[i] + (1 - p0[i]) * qbin
                    # invert in reference: find idx where p0_ref + (1−p0_ref)*quantiles ≈ F
                    # solve q_ref = (F−p0_ref)/(1−p0_ref)
                    q_ref = (
                        (F - p0_ref[i]) / (1 - p0_ref[i])
                        if (1 - p0_ref[i]) > 0
                        else 0.0
                    )
                    # clamp q_ref to [0,1]
                    q_ref = min(max(q_ref, 0.0), 1.0)
                    # choose closest available quantile
                    idx_ref = np.argmin(np.abs(self.quantiles - q_ref))
                    new[i] = Q_ref[i, idx_ref]
            Xcorr[:, f] = new

        # Clamp negatives to zero
        Xcorr[:, len(design_idx) :] = np.where(
            Xcorr[:, len(design_idx) :] < 0, 0, Xcorr[:, len(design_idx) :]
        )

        # rebuild DataFrame
        df_out = pd.DataFrame(
            Xcorr.astype(int), index=df.index, columns=self._col_order
        )
        # Drop non-counts columns
        df_out = df_out.drop([c for c in self.batch_cols], axis=1)
        df_out = df_out.drop([c for c in self.covariate_cols], axis=1)
        df_out = df_out.drop(
            [c for c in df.columns if c not in df_base.columns], axis=1
        )
        other = [c for c in df.columns if c not in self._col_order]
        return pd.concat(
            [
                df_base[other],
                df_base[self.batch_cols],
                df_base[self.covariate_cols],
                df_out,
            ],
            axis=1,
        )


def correctConQuR(
    df,
    batch_cols,
    covariate_cols,
    reference_batch=None,
    quantiles=(0.05, 0.25, 0.5, 0.75, 0.95),
    logistic_kwargs={"penalty": "l2", "solver": "lbfgs", "max_iter": 200},
    quantile_kwargs={"alpha": 0.0},
):
    """
    Conditional logistic quantile regression (ConQuR) for batch correction.

    Parameters
    ----------
    df : pandas.DataFrame
        Input data containing OTU counts and metadata.
    batch_cols : list of str
        List of batch column names.
    covariate_cols : list of str
        List of covariate column names.
    reference_batch : dict, optional
        Dictionary specifying reference batch values for each batch column.
        If None, uses zeros for all batch columns.
    quantiles : tuple of float, optional
        Quantiles to use for quantile regression, by default (0.05, 0.25, 0.5, 0.75, 0.95).
    logistic_kwargs : dict, optional
        Keyword arguments for LogisticRegression.
    quantile_kwargs : dict, optional
        Keyword arguments for QuantileRegressor.

    Returns
    -------
    pandas.DataFrame
        Batch-corrected DataFrame.
    """
    # Define reference batch
    if reference_batch is None:
        reference_batch = {
            batch: ref
            for batch, ref in zip(
                batch_cols, np.zeros(len(batch_cols), dtype=int), strict=False
            )
        }
    # Create model
    conq = ConQur(
        batch_cols=batch_cols,
        covariate_cols=covariate_cols,
        reference_batch=reference_batch,
        quantiles=quantiles,
        logistic_kwargs=logistic_kwargs,
        quantile_kwargs=quantile_kwargs,
    )
    # Fit data into model
    conq.fit(df)
    # Correction
    df_corrected = conq.transform(df)

    return df_corrected


# ComBat-seq
def correctCombatSeq(
    data,
    sample_label,
    batch_label,
    condition_label,
    ref_batch=None,
):
    """
    Perform ComBat-seq batch correction for count data.

    Parameters
    ----------
    data : pandas.DataFrame
        Input data containing count data and metadata.
    sample_label : str
        Column name for sample identifiers.
    batch_label : str
        Column name for batch identifiers.
    condition_label : str
        Column name for condition/experiment identifiers.
    ref_batch : str or None, optional
        Reference batch to use, by default None.

    Returns
    -------
    pandas.DataFrame
        DataFrame with sample, batch, condition, and ComBat-seq corrected counts.
    """
    count_data = data.select_dtypes(include="number")
    batch_data = [batch for batch in data[batch_label]]
    cov_data = [exp for exp in data[condition_label]]

    corrected = pycombat_seq(
        count_data.T, batch_data, covar_mod=cov_data, ref_batch=ref_batch
    )
    corrected_df = pd.DataFrame(
        corrected.T, index=data.index, columns=count_data.columns
    )

    return pd.concat(
        [data[sample_label], data[batch_label], data[condition_label], corrected_df],
        axis=1,
    )
